 ##############################################################################
#
# Action Replier
# v1.4 (27/03/2016)   2013-2016 Menz Agitat
# d'aprs une ide originale de mindcry
# avec son script Action Reverse Replier (2007 mindcry)
#
# IRC: irc.epiknet.org  #boulets / #eggdrop
#
# Mes scripts sont tlchargeables sur http://www.eggdrop.fr
# Retrouvez aussi toute l'actualit de mes releases sur
# http://www.boulets.oqp.me/tcl/scripts/index.html
#
 ##############################################################################

#
# Description
#
# Ce script permet  votre eggdrop de rpondre lorsqu'un CTCP ACTION (un /me)
# l'implique.
# Par dfaut, il renversera la situation en substituant son nick par celui de
# la personne qui a fait le /me et vice versa. Nanmoins, il est aussi capable
# d'apprendre et de rutiliser les actions qu'il aura apprises.
#
# Lorsqu'une action est apprise, les nicks des personnes prsentes sur le chan
# sont automatiquement substitus par des %-%randnick%-%, qui seront  nouveau
# substitus par des nicks de personnes prsentes au moment de s'en resservir.
# De mme, le nick de l'acteur est substitu par %-%actor%-% et le nick du bot
# par %-%botnick%-%, afin de permettre aux actions mmorises de s'adapter au
# contexte.
#
# Voici quelques exemples dmontrant la faon dont il renverse les actions,
#	les apprend puis s'en ressert :
#		* MenzAgitat apprend  L_Eggdrop des trucs que seul MenzAgitat sait
#		* L_Eggdrop apprend  MenzAgitat des trucs que seul L_Eggdrop sait
#		* MenzAgitat tape L_Eggdrop
#		* L_Eggdrop tape MenzAgitat
#		* MenzAgitat cherche des bugs dans L_Eggdrop
#		* L_Eggdrop tape MenzAgitat
#
# Action Replier peut optionnellement fonctionner en collaboration avec
# l'Interface MegaHAL (script du mme auteur  tlcharger sparment) afin
# d'viter les rponses doubles et de ne pas rpondre si le mode "TAGUEULE"
# de MegaHAL est activ. Notez que dans ce cas, vous devrez charger
# le script Interface MegaHAL avant Action Replier.
#
# La commande !forgetaction vous permet d'effacer une action apprise.
#
# La commande !actionreplier_stats permet au propritaire de l'eggdrop
# d'afficher le nombre d'actions stockes dans la base de donnes.
#
# Pour activer Action Replier sur un chan, vous devez taper ceci en partyline de
# l'eggdrop :
# 	.chanset #NomDuChan +actionreplier
# et ceci pour le dsactiver :
# 	.chanset #NomDuChan -actionreplier
#
# Pour activer l'apprentissage des actions sur un chan, vous devez taper ceci en
# partyline de l'eggdrop :
# 	.chanset #NomDuChan +actionreplier_canlearn
# et ceci pour le dsactiver :
# 	.chanset #NomDuChan -actionreplier_canlearn
# Il est ncessaire d'activer Action Replier si vous voulez qu'il apprenne (voir
# le paragraphe prcdent).
#

#
# Changelog
#
# 1.0
#		- 1re version
# 1.1
#		- Ajout de l'option exclusion_list vous permettant de spcifier des nicks
#			ou handles d'utilisateurs auxquels Action Replier ne rpondra jamais.
# 1.2
#		- Correction : dsormais, Action Replier ragit mme si la casse des
#			caractres n'est pas respecte dans le nick de l'eggdrop. 
#		- Correction : le remplacement du nick de l'eggdrop par %-%botnick%-% ne
#			fonctionnait pas correctement si son nick tait rpt plusieurs fois
#			d'affile.
#		- Quelques corrections et ajustements mineurs.
# 1.3
#		- Ajout de l'option ignore_flags permettant de dfinir des autorisations
#			rdhibitoires.
# 1.4
#		- Correction de la faute d'orthographe dans le nom du script (a s'crit
#			"replier" et non pas "replyer"). Attention, le flag actionreplier a
#			chang de nom lui aussi, ainsi que le fichier de la base de donnes.
#		- Modification : le script ne ragira plus sur le nick de l'Eggdrop s'il est
#			prcd d'un dise, afin d'viter les tentatives de spam.
# 

#
# Licence
#
#		Cette cration est mise  disposition selon le Contrat
#		Attribution-NonCommercial-ShareAlike 3.0 Unported disponible en ligne
#		http://creativecommons.org/licenses/by-nc-sa/3.0/ ou par courrier postal 
#		Creative Commons, 171 Second Street, Suite 300, San Francisco, California
#		94105, USA.
#		Vous pouvez galement consulter la version franaise ici :
#		http://creativecommons.org/licenses/by-nc-sa/3.0/deed.fr
#

if {[info commands ::actionreplier::uninstall] eq "::actionreplier::uninstall"} { ::actionreplier::uninstall }
if { [join [split [::tcl::string::range [lindex $version 0] 0 5] "."] ""] < 1620 } { putloglev o * "\00304\002\[Action Replier - Erreur\]\002\003 La version de votre eggdrop est \00304\002$version\002\003; Action Replier ne fonctionnera correctement que sur les eggdrops version 1.6.20 ou suprieure." ; return }
if { [info tclversion] < 8.5 } { putloglev o * "\00304\002\[Action Replier - Erreur\]\002\003 Action Replier ncessite que Tcl 8.5 (ou plus) soit install pour fonctionner. Votre version actuelle de Tcl est \00304\002$tcl_version\002\003." ; return }
package require Tcl 8.5
namespace eval ::actionreplier {



 ############################################################################
### Configuration
 ############################################################################

	# Liste de nicks ou handles d'utilisateurs qu'Action Replier doit ignorer.
	variable exclusion_list {}

	# Autorisations rdhibitoires (Action Replier ignorera les utilisateurs
	# portant au moins l'un de ces flags.
	# Si vous ne souhaitez pas utiliser cette option, mettez "".
	# Si par exemple vous souhaitez dfinir "v" en autorisations locales et ne pas
	# dfinir d'autorisations globales, ne mettez surtout pas "-|v", car Action
	# Replier ignorerait tout le monde. Mettez plutt "|v".
	variable ignore_flags ""

	# Autoriser Action Replier  rpondre aux messages privs ? (0 = non, 1 = oui)
	variable allow_private_msgs 0

	# Autoriser Action Replier  apprendre les actions dans les messages privs ?
	# (0 = non, 1 = oui)
	variable can_learn_priv_msgs 0

	# Frquence (en %)  laquelle le bot va choisir une rponse au hasard dans la
	# base de donnes d'actions qu'il a apprises, plutt que de retourner
	# simplement l'action inverse.
	variable database_use_rate 57

	# Chemin et nom du fichier de la base de donnes
	variable database_file "scripts/ActionReplier.dat"

	# Commande utilise pour effacer une action mmorise
	variable forgetaction_cmd "!forgetaction"

	# Autorisations requises pour pouvoir effacer une action mmorise
	variable forgetaction_auth "m|m"

	# Anti-flood (0 = dsactiv, 1 = activ)
	# Veuillez noter que l'anti-flood ne signale pas lorsqu'il s'active, Action
	# Replier cesse simplement de rpondre.
	variable antiflood 1

	# Combien de rponses aux actions sont autorises en combien de temps ?
	# exemple : "10:120" = 10 rponses maximum en 120 secondes;
	# les suivantes seront ignores.
	variable flood_limit "10:120"
	
	# Ce script est-il destin  fonctionner en collaboration avec l'Interface
	# MegaHAL ? (script du mme auteur  tlcharger sparment)
	# (0 = non ; 1 = oui)
	variable megahal_interface 0

 ############################################################################
### Fin de la configuration
 ############################################################################




	 ##############################################################################
	### Initialisation
	 ##############################################################################
  variable scriptname "Action Replier"
  variable version "1.31.20130202"
	if { $megahal_interface } {
		# dtection du package MegaHAL_Interface
		if {
			([lsearch [package names] "MegaHAL_Interface"] == -1)
			|| ([package vcompare [package present MegaHAL_Interface] 3.4.0] == -1)
		} then {
			putloglev o * "\00304\002\[[set ::actionreplier::scriptname] - ERREUR\]\002\003 Vous avez activ l'option \megahal_interface\002 mais le script \002Interface MegaHAL\002 v3.4.0 (ou plus) n'est apparemment pas charg."
			namespace delete ::actionreplier
			return
		}
		after [expr ($megahal_interface::IA_activation_delay + 2) * 1000] { ::actionreplier::rebind }
	}
	setudef flag actionreplier
	setudef flag actionreplier_canlearn
	regexp {(.*)\|(.*)} $::actionreplier::ignore_flags {} ::actionreplier::ignore_global_flags ::actionreplier::ignore_local_flags
	# initialisation de l'antiflood
	scan $flood_limit "%d:%d" ::actionreplier::max_instances ::actionreplier::instance_length
	array set ::actionreplier::instance {}
	# lecture de la base de donnes
	if { [file exists $::actionreplier::database_file] } {
		set databasefile [open $::actionreplier::database_file r]
		set ::actionreplier::database [read -nonewline $databasefile]
		close $databasefile
	} else {
		set ::actionreplier::database {}
	}
	# Procdure de dsinstallation (le script se dsinstalle totalement avant chaque rehash ou  chaque relecture au moyen de la commande "source" ou autre)
	proc uninstall {args} {
		putlog "Dsallocation des ressources de \002$::actionreplier::scriptname\002..."
		foreach binding [lsearch -inline -all -regexp [binds *[set ns [::tcl::string::range [namespace current] 2 end]]*] " \{?(::)?$ns"] {
			unbind [lindex $binding 0] [lindex $binding 1] [lindex $binding 2] [lindex $binding 4]
		}
		foreach running_utimer [utimers] {
			if { [::tcl::string::match "*[namespace current]::*" [lindex $running_utimer 1]] } { killutimer [lindex $running_utimer 2] }
		}
		if { $::actionreplier::megahal_interface } { bind ctcp - ACTION ::megahal_interface::process_ctcp_action }
		namespace delete ::actionreplier
	}
}

 ##############################################################################
### Traitement des CTCP ACTION
 ##############################################################################
proc ::actionreplier::process_ctcp_action {nick host hand target ctcp_type text} {
	set hash [md5 $target]
	if { [validchan $target] } {
		regexp {(.*)\s(.*)} [split [chattr $hand $target] "|"] {} user_global_flags user_local_flags
	} else {
		set user_global_flags [chattr $hand]
		set user_local_flags ""
	}
	if { $hand eq "*" } {
		set user_local_flags "-"
		set user_global_flags "-"
	}
	if { (([validchan $target])
		&& (![channel get $target actionreplier]))
		|| ((![validchan $target])
		&& (!$::actionreplier::allow_private_msgs))
		|| (![regexp -nocase "(^|\[^a-z#\])[regsub -all {\W} $::botnick {\\&}](\[^a-z\]|$)" $text])
		|| (($actionreplier::antiflood == 1)
		&& ([::actionreplier::antiflood $hash]))
		|| ([lsearch -nocase $::actionreplier::exclusion_list $nick] != -1)
		|| ([lsearch -nocase -exact $::actionreplier::exclusion_list $hand] != -1)
		|| (($::actionreplier::ignore_flags ne "")
		&& ((($::actionreplier::ignore_global_flags ne "")
		&& ([regexp "\[$::actionreplier::ignore_global_flags\]" "-$user_global_flags"]))
		|| (($::actionreplier::ignore_local_flags ne "")
		&& ([regexp "\[$::actionreplier::ignore_local_flags\]" "-$user_local_flags"]))))
		|| (($::actionreplier::megahal_interface)
		&& ([::tcl::info::exists ::megahal_interface::shutup([set clean_chan [::megahal_interface::neutralize_special_chars $target]])])
		&& ($::megahal_interface::shutup($clean_chan)))
	} then {
		::actionreplier::antiflood_close_instance $hash
		if { $::actionreplier::megahal_interface } {
			::megahal_interface::process_ctcp_action $nick $host $hand $target $ctcp_type $text
		}
		return
	} else {
		set split_text [split $text]
		# on a $database_use_rate% de chances pour que la rponse soit pioche dans la base de donnes
		if {
			($::actionreplier::database ne "")
			&& ([expr ([clock clicks -milliseconds] % 100) + 1] <= $::actionreplier::database_use_rate)
		} then {
			set reply [::actionreplier::get_action_from_database $nick $target]
		} else {
			set reply $text
		}
		# on substitue les nicks en vue d'apprendre l'action
		if { [validchan $target] } {
			set randnicked_split_text [::actionreplier::substitute_nicks $nick $target $split_text]
		} else {
			set randnicked_split_text $split_text
		}
		# si l'apprentissage est activ et qu'on ne connat pas dj cette action, on l'ajoute  la base de donnes
		if {
			((([validchan $target])
			&& ([channel get $target actionreplier_canlearn]))
			|| ((![validchan $target])
			&& ($::actionreplier::can_learn_priv_msgs)))
			&& ([lsearch -exact $::actionreplier::database $randnicked_split_text] == -1)
		} then {
			::actionreplier::store_action_in_database $nick $target $randnicked_split_text
		}
		# on substitue le nick de l'eggdrop par le nick de la personne qui a fait le /me et inversement
		regsub -all {\s+} [regsub -all -nocase "(^|\[^a-z\])[regsub -all {\W} [set ::botnick] {\\&}](\[^a-z\]|$)" [regsub -all {\s} $reply "  "] "\\1!-!-!\\2"] " " reply
		regsub -all {\s+} [regsub -all -nocase "(^|\[^a-z\])[regsub -all {\W} [set nick] {\\&}](\[^a-z\]|$)" [regsub -all {\s} $reply "  "] "\\1[set ::botnick]\\2"] " " reply
		regsub -all "!-!-!" $reply $nick reply
		if { $target eq $::botnick } {
			puthelp "PRIVMSG $nick :\001ACTION $reply\001"
		} else {
			puthelp "PRIVMSG $target :\001ACTION $reply\001"
			putloglev p $target "Action: $::botnick $reply"
		}
	}
}

 ##############################################################################
### Ajoute une action  la base de donnes (en mmoire ET sur le disque)
 ##############################################################################
proc ::actionreplier::store_action_in_database {nick target text} {
	lappend ::actionreplier::database $text
	set databasefile [open $::actionreplier::database_file a]
	puts $databasefile "{[set text]}"
	close $databasefile
}

 ##############################################################################
### Retourne une action choisie alatoirement dans la base de donnes
 ##############################################################################
proc ::actionreplier::get_action_from_database {nick target} {
	set reply [join [lindex $::actionreplier::database [expr ([clock clicks -milliseconds] % [llength $::actionreplier::database])]]]
	regsub -all {%-%botnick%-%} $reply $::botnick reply
	regsub -all {%-%actor%-%} $reply $nick reply
	if { [validchan $target] } {
		while { [::tcl::string::match "*%-%randnick%-%*" $reply] } {
			regsub "%-%randnick%-%" $reply [::actionreplier::randnick $nick $target] reply
		}
		return $reply
	} else {
		return [regsub -all "%-%randnick%-%" $reply "quelqu'un"]
	}
}

 ##############################################################################
### Substitue le nick de l'auteur du /me par %-%actor%-%, les nicks des autres
### utilisateurs prsents par %-%randnick%-%, et le nick de l'eggdrop par
### %-%botnick%-% lors de l'apprentissage.
 ##############################################################################
proc ::actionreplier::substitute_nicks {nick chan split_text} {
	regsub -all {\s+} [regsub -all -nocase "(^|\[^a-z\])[regsub -all {\W} $::botnick {\\&}](\[^a-z\]|$)" [regsub -all {\s} $split_text "  "] "\\1%-%botnick%-%\\2"] " " split_text
	regsub -all {\s+} [regsub -all -nocase "(^|\[^a-z\])[regsub -all {\W} $nick {\\&}](\[^a-z\]|$)" [regsub -all {\s} $split_text "  "] "\\1%-%actor%-%\\2"] " " split_text
	foreach current_nick [chanlist $chan] {
		regsub -all {\s+} [regsub -all -nocase "(^|\[^a-z\])[regsub -all {\W} $current_nick {\\&}](\[^a-z\]|$)" [regsub -all {\s} $split_text "  "] "\\1%-%randnick%-%\\2"] " " split_text
	}
	return $split_text
}

 ##############################################################################
### Retourne un nick alatoire parmi les users prsents sur le chan
### (sauf l'eggdrop, sauf les autres eggs du botnet, sauf l'user qui a pos la
### question)
 ##############################################################################
proc ::actionreplier::randnick {nick chan} {
	set users_list [lreplace [set users [chanlist $chan -b|]] [set index [lsearch $users $nick]] $index]
	set users_list [lreplace $users_list [set index [lsearch $users_list $::botnick]] $index]
	if {![llength $users_list]} {
		return "quelqu'un"
	}
	return [lindex $users_list [rand [llength $users_list]]]
}

 ##############################################################################
### Contrle du flood
 ##############################################################################
proc ::actionreplier::antiflood {hash} {
	if {![info exists ::actionreplier::instance($hash)]} {
		set ::actionreplier::instance($hash) 0
	}
	if { $::actionreplier::instance($hash) >= $::actionreplier::max_instances } {
		return 1
	} else {
		incr ::actionreplier::instance($hash)
		utimer $::actionreplier::instance_length [list ::actionreplier::antiflood_close_instance $hash]
		return 0
	}
}
proc ::actionreplier::antiflood_close_instance {hash} {
	if {
		([info exists ::actionreplier::instance($hash)])
		&& ($::actionreplier::instance($hash) > 0)
	} then {
		incr ::actionreplier::instance($hash) -1
	}
}

 ##############################################################################
### Reprise de contrle du bind ctcp de l'Interface MegaHAL
 ##############################################################################
proc ::actionreplier::on_connect {args} {
	after [expr ($megahal_interface::IA_activation_delay + 2) * 1000] { ::actionreplier::rebind }
}
proc ::actionreplier::rebind {} {
	if { [binds ::megahal_interface::process_ctcp_action] ne "" } {
		unbind ctcp - ACTION ::megahal_interface::process_ctcp_action
	}
}

 ##############################################################################
### Effacement d'une action apprise
 ##############################################################################
proc ::actionreplier::action_deletion {nick host hand chan text} {
	if { $text eq "" } {
		puthelp "PRIVMSG $chan :\037Syntaxe\037 : \002[set ::actionreplier::forgetaction_cmd]\002 \00314<\003action  oublier\00314>\003 \00307|\003 Efface une action mmorise."
	} else {
		set text [::tcl::string::range $text [expr [::tcl::string::length $::botnick] + 1] end]
		set split_text [split $text]
		foreach current_nick [chanlist $chan] {
			regsub "(^|\[^a-z\])[regsub -all {\W} $current_nick {\\&}](\[^a-z\]|$)" $split_text "\\1%-%nick%-%\\2" split_text
		}
		set temp_database [::tcl::string::map {%-%botnick%-% %-%nick%-% %-%actor%-% %-%nick%-% %-%randnick%-% %-%nick%-%} $::actionreplier::database]
		if { [set action_index [lsearch -exact $temp_database $split_text]] == -1 } {
			puthelp "PRIVMSG $chan :\00314Cette action n'a pas t trouve dans la base de donnes.\003"
		} else {
			set deleted_action [lindex $::actionreplier::database $action_index]
			set ::actionreplier::database [lreplace $::actionreplier::database $action_index $action_index]
			# criture de la base de donnes
			set databasefile [open $::actionreplier::database_file w]
			foreach action $::actionreplier::database {
				puts $databasefile "{[set action]}"
			}
			close $databasefile
			puthelp "PRIVMSG $chan :\00314L'action \00307[join $deleted_action]\00314 a t oublie."
		}
	}
}

 ##############################################################################
### Commande !actionreplier_stats : affiche le nombre d'actions stockes dans
### la base de donnes
 ##############################################################################
proc ::actionreplier::stats {nick host hand chan arg} {
	puthelp "PRIVMSG $chan :\00314\[[set ::actionreplier::scriptname]\]\003 \002[llength $::actionreplier::database]\002 actions dans la base de donnes."
}

 ##############################################################################
### Binds
 ##############################################################################
bind evnt - prerehash ::actionreplier::uninstall
bind ctcp - ACTION ::actionreplier::process_ctcp_action
bind pub $::actionreplier::forgetaction_auth $::actionreplier::forgetaction_cmd ::actionreplier::action_deletion
bind pub n !actionreplier_stats ::actionreplier::stats
if { $::actionreplier::megahal_interface } {
	bind evnt - init-server ::actionreplier::on_connect
}


putlog "[set ::actionreplier::scriptname] v[set ::actionreplier::version] (2013-2016 Menz Agitat) a t charg."
